#pragma once

#include "libsyntax_global.h"
#include <libtoken.h>
#include <memory>
#include <tuple>
#include <functional>

namespace ast_basic {
    class TokenNode;
    class ExprNode;
}

namespace lib_syntax {
    /**
     * @brief ﷨쳣
     */
    class SyntaxException {
    private:
        QString msg_store;

    public:
        SyntaxException(const QString& message);
        virtual ~SyntaxException() = default;

        virtual QString message() const;
    };

    // ﷨ӿ ===================================================================================================
    /**
     * @brief ﷨ƥӿ
     */
    class BaseRule {
    public:
        /**
         * @brief ӹ
         * @return
         */
        virtual QList<std::shared_ptr<const BaseRule>> children() const = 0;

        /**
         * @brief ƥ
         */
        enum class MatchResult {
            Success, // ƥ
            Part,    // ƥ
            Fail     // ӵһȫƥ
        };

        /**
         * @brief tokenƥ
         * @return ײ룬ƥtoken
         */
        virtual std::tuple<MatchResult, uint> match(const QList<std::shared_ptr<const lib_token::Token>>& stream) const = 0;

        /**
         * @brief 
         * @param stream
         * @return
         */
        virtual QList<std::shared_ptr<const ast_basic::TokenNode>> parse(const QList<std::shared_ptr<const lib_token::Token>>& stream) const = 0;

        /**
         * ƥ﷨Ĵʷб
         *
         * \return ʷ
         */
        virtual QString token_present() const = 0;
    };

    /**
     * @brief tokenƥ
     */
    class LIBSYNTAX_EXPORT TokenMatch : public BaseRule {
    private:
        std::shared_ptr<const lib_token::TokenDefine> define_peer;

    public:
        TokenMatch(std::shared_ptr<const lib_token::TokenDefine> define);

        // BaseRule interface
    public:
        virtual QList<std::shared_ptr<const BaseRule>> children() const override;
        virtual std::tuple<MatchResult, uint> match(const QList<std::shared_ptr<const lib_token::Token>>& stream) const override;
        virtual QList<std::shared_ptr<const ast_basic::TokenNode>> parse(const QList<std::shared_ptr<const lib_token::Token>>& stream) const override;
        virtual QString token_present() const override;
    };

    /**
     * @brief ﷨ƥ
     */
    class LIBSYNTAX_EXPORT Any : public BaseRule {
    private:
        QList<std::shared_ptr<const BaseRule>> mbrs_store;

        std::tuple<MatchResult, uint, std::shared_ptr<const BaseRule>> rule_select(const QList<std::shared_ptr<const lib_token::Token>>& stream) const;

    public:
        Any(const QList<std::shared_ptr<const BaseRule>> mbrs);

        // BaseRule interface
    public:
        virtual QList<std::shared_ptr<const BaseRule>> children() const override;
        virtual std::tuple<MatchResult, uint> match(const QList<std::shared_ptr<const lib_token::Token>>& stream) const override;
        virtual QList<std::shared_ptr<const ast_basic::TokenNode>> parse(const QList<std::shared_ptr<const lib_token::Token>>& stream) const override;
        virtual QString token_present() const override;
    };

    /**
     * @brief ﷨ƥ
     */
    class LIBSYNTAX_EXPORT Seqs : public BaseRule {
    private:
        QList<std::shared_ptr<const BaseRule>> mbrs_store;

    public:
        Seqs(const QList<std::shared_ptr<const BaseRule>> mbrs);

        // BaseRule interface
    public:
        virtual QList<std::shared_ptr<const BaseRule>> children() const override;
        virtual std::tuple<MatchResult, uint> match(const QList<std::shared_ptr<const lib_token::Token>>& stream) const override;
        virtual QList<std::shared_ptr<const ast_basic::TokenNode>> parse(const QList<std::shared_ptr<const lib_token::Token>>& stream) const override;
        virtual QString token_present() const override;
    };

    /**
     * @brief ﷨ظƥ
     */
    class LIBSYNTAX_EXPORT Rept : public BaseRule {
    private:
        std::shared_ptr<const BaseRule> rule_peer;
        int min_match, max_match;

    public:
        Rept(std::shared_ptr<const BaseRule> rule, int min, int max);

        // BaseRule interface
    public:
        virtual QList<std::shared_ptr<const BaseRule>> children() const override;
        virtual std::tuple<MatchResult, uint> match(const QList<std::shared_ptr<const lib_token::Token>>& stream) const override;
        virtual QList<std::shared_ptr<const ast_basic::TokenNode>> parse(const QList<std::shared_ptr<const lib_token::Token>>& stream) const override;
        virtual QString token_present() const override;
    };

    // ﷨ʵ ===================================================================================================
    /**
     * @brief Ӧ﷨ʽ
     */
    class LIBSYNTAX_EXPORT ExprRule : public lib_syntax::BaseRule, public std::enable_shared_from_this<ExprRule> {
    private:
        std::shared_ptr<const lib_syntax::BaseRule> child_store;
        QString name_store;
        int mark_store;
        std::function<QList<std::shared_ptr<const ast_basic::TokenNode>>(const QList<std::shared_ptr<const ast_basic::TokenNode>>& primary)> filter_proc;

    public:
        typedef QList<std::shared_ptr<const ast_basic::TokenNode>> TokenSeqs;
        ExprRule(const QString& rule_name, int expr_mark);

        virtual std::shared_ptr<const ExprRule> reloadRule(std::function<TokenSeqs(const TokenSeqs&)> filter, std::shared_ptr<const BaseRule> rule);
        virtual QString name() const;
        virtual int typeMark() const;

        // BaseRule interface
    public:
        virtual QList<std::shared_ptr<const lib_syntax::BaseRule>> children() const override;
        virtual std::tuple<MatchResult, uint> match(const QList<std::shared_ptr<const lib_token::Token>>& stream) const override;
        virtual QList<std::shared_ptr<const ast_basic::TokenNode>> parse(const QList<std::shared_ptr<const lib_token::Token>>& stream) const override;
        virtual QString token_present() const override;
    };

    /**
     * @brief ﷨˶Թ
     */
    class LIBSYNTAX_EXPORT ExprsChecker {
    public:
        ExprsChecker(std::shared_ptr<const ExprRule> parse_tree);

        QList<std::shared_ptr<const ast_basic::TokenNode>> parseFrom(const QList<std::shared_ptr<const lib_token::Token>>& all_tokens);

    private:
        std::shared_ptr<const ExprRule> syntax_tree_root;
    };

} // namespace lib_syntax